from dataclasses import dataclass

import pint

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

from .shielding_gas_type import ShieldingGasType

__all__ = ["ShieldingGasForProcedure", "ShieldingGasForProcedureType"]


@dataclass
class ShieldingGasForProcedure:
    """<CLASS DOCSTRING>"""

    use_torch_shielding_gas: bool
    torch_shielding_gas: ShieldingGasType
    torch_shielding_gas_flowrate: pint.Quantity
    use_backing_gas: bool = None
    backing_gas: ShieldingGasType = None
    backing_gas_flowrate: pint.Quantity = None
    use_trailing_gas: bool = None
    trailing_shielding_gas: ShieldingGasType = None
    trailing_shielding_gas_flowrate: pint.Quantity = None


class ShieldingGasForProcedureType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/process/shielding_gas_for_procedure"
    version = "1.0.0"
    types = [ShieldingGasForProcedure]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = ShieldingGasForProcedure(**tree)
        return obj



from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.measurement import Error

__all__ = ["Error", "ErrorType"]


class ErrorType(WeldxType):
    """Serialization class for measurement errors."""

    name = "measurement/error"
    version = "1.0.0"
    types = [Error]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: Error, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = Error(**tree)
        return obj



from .datetimeindex import DatetimeIndexType
from .timedelta import TimedeltaType
from .timedeltaindex import TimedeltaIndexType
from .timestamp import TimestampType

__all__ = ["TimestampType", "DatetimeIndexType", "TimedeltaType", "TimedeltaIndexType"]



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Tuple

from pandas import Timestamp

from weldx.asdf.types import WeldxType
from weldx.transformations import CoordinateSystemManager, LocalCoordinateSystem


@dataclass
class CoordinateTransformation:
    """Stores data of a coordinate transformation."""

    name: str
    reference_system: str
    transformation: LocalCoordinateSystem


class CoordinateTransformationASDF(WeldxType):
    """Serialization class for CoordinateTransformation"""

    name = "core/transformations/coordinate_transformation"
    version = "1.0.0"
    types = [CoordinateTransformation]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {}

    @classmethod
    def to_tree(cls, node: CoordinateTransformation, ctx):
        """
        Convert a 'CoordinateTransformation' instance into YAML representations.

        Parameters
        ----------
        node :
            Instance of the 'CoordinateTransformation' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'CoordinateTransformation'
            type to be serialized.

        """
        tree = {
            "name": node.name,
            "reference_system": node.reference_system,
            "transformation": node.transformation,
        }
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into a 'CoordinateTransformation'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        CoordinateTransformation :
            An instance of the 'CoordinateTransformation' type.

        """
        return CoordinateTransformation(
            name=tree["name"],
            reference_system=tree["reference_system"],
            transformation=tree["transformation"],
        )


@dataclass
class CoordinateSystemManagerSubsystem:
    """Helper class to collect all relevant data of a CSM subsystem."""

    name: str
    parent_system: str
    reference_time: Timestamp
    root_cs: str
    subsystems: List[str]
    members: List[str]


class CoordinateSystemManagerSubsystemASDF(WeldxType):
    """Serialization class for a CoordinateSystemManagerSubsystem instance"""

    name = "core/transformations/coordinate_system_hierarchy_subsystem"
    version = "1.0.0"
    types = [CoordinateSystemManagerSubsystem]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {}

    @classmethod
    def to_tree(cls, node: CoordinateSystemManagerSubsystem, ctx):
        """Convert a 'CoordinateSystemManagerSubsystem' instance into YAML repr.

        Parameters
        ----------
        node :
            Instance of the 'CoordinateSystemManagerSubsystem' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the
            'CoordinateSystemManagerSubsystem' type to be serialized.

        """
        tree = {
            "name": node.name,
            "root_cs": node.root_cs,
            "reference_time": node.reference_time,
            "parent_system": node.parent_system,
            "subsystem_names": node.subsystems,
            "members": node.members,
        }
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts YAML trees into a 'CoordinateSystemManagerSubsystem'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        CoordinateSystemManagerSubsystem :
            An instance of the 'CoordinateSystemManagerSubsystem
        """
        return tree


class CoordinateSystemManagerASDF(WeldxType):
    """Serialization class for weldx.transformations.LocalCoordinateSystem"""

    name = "core/transformations/coordinate_system_hierarchy"
    version = "1.0.0"
    types = [CoordinateSystemManager]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {}

    @classmethod
    def _extract_all_subsystems(
        cls, csm: CoordinateSystemManager, _recursive_call: bool = False
    ) -> List[Tuple[CoordinateSystemManager, str]]:
        """Return a list of all subsystems and their corresponding parent names.

        This function extracts all subsystems and nested subsystems of the passed
        coordinate system manager instance. Each subsystem is stored inside a tuple
        together with the parent systems' name.
        Parameters
        ----------
        csm:
            Coordinate system manager instance
        _recursive_call:
            Specifies if the current function call happened recursively

        Returns
        -------
        List[Tuple[CoordinateSystemManager, str]]:
            List of subsystems and their parent names

        """
        if not _recursive_call:
            csm = deepcopy(csm)

        subsystems = csm.unmerge()

        subsystems = [(subsystem, csm.name) for subsystem in subsystems]
        for subsystem, _ in subsystems:
            if subsystem.number_of_subsystems > 0:
                subsystems += cls._extract_all_subsystems(subsystem, True)

        return subsystems

    @classmethod
    def _extract_subsystem_data(
        cls, csm: CoordinateSystemManager
    ) -> List[CoordinateSystemManagerSubsystem]:
        """Get the subsystem data of a CoordinateSystemManager instance.

        Parameters
        ----------
        csm:
            CoordinateSystemManager instance.

        Returns
        -------
        List[CoordinateSystemManagerSubsystem]:
            Data of all subsystems

        """
        subsystems = cls._extract_all_subsystems(csm)
        subsystem_data = []
        for subsystem, parent in subsystems:
            child_systems = [
                child.name for child, parent in subsystems if parent == subsystem.name
            ]
            subsystem_data += [
                CoordinateSystemManagerSubsystem(
                    subsystem.name,
                    parent,
                    subsystem.reference_time,
                    subsystem.root_system_name,
                    child_systems,
                    subsystem.coordinate_system_names,
                )
            ]
        return subsystem_data

    @classmethod
    def _merge_subsystems(cls, tree, csm, subsystem_data_list):
        """Merge all subsystems into the CoordinateSystemManager instance.

        Parameters
        ----------
        tree:
            The dictionary representing the ASDF files YAML tree
        csm:
            CoordinateSystemManager instance
        subsystem_data_list:
            List containing all relevant subsystem data from the CoordinateSystemManager
            instance

        """
        subsystem_data_dict = {
            subsystem_data["name"]: subsystem_data
            for subsystem_data in subsystem_data_list
        }

        if subsystem_data_list:
            cls._recursively_merge_subsystems(
                csm, tree["subsystem_names"], subsystem_data_dict
            )

    @classmethod
    def _recursively_merge_subsystems(
        cls, csm, subsystem_names, subsystem_data_dict: Dict
    ):
        """Merge a list of subsystems into a CoordinateSystemManager instance.

        This function also considers nested subsystems using recursive function calls.

        Parameters
        ----------
        csm:
            CoordinateSystemManager instance
        subsystem_names:
            Names of all subsystems that should be added.
        subsystem_data_dict
            Dictionary containing the data of all subsystems from the
            CoordinateSystemManager instance

        """
        for subsystem_name in subsystem_names:
            subsystem_data = subsystem_data_dict[subsystem_name]
            if subsystem_data["subsystem_names"]:
                cls._recursively_merge_subsystems(
                    subsystem_data["csm"],
                    subsystem_data["subsystem_names"],
                    subsystem_data_dict,
                )
            csm.merge(subsystem_data["csm"])

    @classmethod
    def _add_coordinate_systems_to_subsystems(
        cls, tree, csm, subsystem_data_list: List[Dict]
    ):
        """Add all coordinate systems to the owning subsystem.

        Parameters
        ----------
        tree:
            The dictionary representing the ASDF files YAML tree
        csm:
            CoordinateSystemManager instance
        subsystem_data_list:
            A list containing all relevant subsystem data

        """
        main_system_lcs = []

        for lcs_data in tree["coordinate_systems"]:
            edge = [lcs_data.name, lcs_data.reference_system]
            is_subsystem_lcs = False
            for subsystem_data in subsystem_data_list:
                if set(edge).issubset(subsystem_data["members"]):
                    subsystem_data["lcs"] += [(*edge, lcs_data.transformation)]
                    is_subsystem_lcs = True
                    break
            if not is_subsystem_lcs:
                main_system_lcs += [(*edge, lcs_data.transformation)]

        # add coordinate systems to corresponding csm
        cls._add_coordinate_systems_to_manager(csm, main_system_lcs)
        for subsystem_data in subsystem_data_list:
            cls._add_coordinate_systems_to_manager(
                subsystem_data["csm"], subsystem_data["lcs"]
            )

    @classmethod
    def _add_coordinate_systems_to_manager(
        cls,
        csm: CoordinateSystemManager,
        lcs_data_list: List[Tuple[str, str, LocalCoordinateSystem]],
    ):
        """Add all coordinate systems to a CSM instance.

        Parameters
        ----------
        csm:
            CoordinateSystemManager instance.
        lcs_data_list:
            List containing all the necessary data of all coordinate systems that should
            be added.

        """
        # todo: ugly but does the job. check if this can be enhanced
        leaf_nodes = [csm.root_system_name]
        while lcs_data_list:
            leaf_nodes_next = []
            lcs_data_list_next = []
            for lcs_data in lcs_data_list:
                lcs_added = False
                for leaf_node in leaf_nodes:
                    if leaf_node in lcs_data[0:2]:
                        if leaf_node == lcs_data[0]:
                            csm.add_cs(lcs_data[1], leaf_node, lcs_data[2], False)
                            leaf_nodes_next += [lcs_data[1]]
                        else:
                            csm.add_cs(lcs_data[0], leaf_node, lcs_data[2], True)
                            leaf_nodes_next += [lcs_data[0]]
                        lcs_added = True
                        break
                if not lcs_added:
                    lcs_data_list_next += [lcs_data]

            leaf_nodes = leaf_nodes_next
            lcs_data_list = lcs_data_list_next

    @classmethod
    def to_tree(cls, node: CoordinateSystemManager, ctx):
        """Convert a 'CoordinateSystemManager' instance into YAML representations.

        Parameters
        ----------
        node :
            Instance of the 'CoordinateSystemManager' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'CoordinateSystemManager'
            type to be serialized.

        """
        graph = deepcopy(node.graph)  # TODO: Check if deepcopy is necessary

        # remove automatically computed edges (inverted directions)
        remove_edges = []
        for edge in graph.edges:
            if not graph.edges[edge]["defined"]:
                remove_edges.append(edge)
        graph.remove_edges_from(remove_edges)

        coordinate_system_data = []
        for name, reference_system in graph.edges:
            coordinate_system_data += [
                CoordinateTransformation(
                    name,
                    reference_system,
                    graph.edges[(name, reference_system)]["lcs"],
                )
            ]

        subsystem_data = cls._extract_subsystem_data(node)
        subsystems = [
            subsystem.name
            for subsystem in subsystem_data
            if subsystem.parent_system == node.name
        ]

        tree = {
            "name": node.name,
            "reference_time": node.reference_time,
            "subsystem_names": subsystems,
            "subsystems": subsystem_data,
            "root_system_name": node.root_system_name,
            "coordinate_systems": coordinate_system_data,
        }
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into a 'CoordinateSystemManager'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        CoordinateSystemManager :
            An instance of the 'CoordinateSystemManager' type.

        """
        reference_time = None
        if "reference_time" in tree:
            reference_time = tree["reference_time"]
        csm = CoordinateSystemManager(
            tree["root_system_name"], tree["name"], time_ref=reference_time
        )

        subsystem_data_list = tree["subsystems"]

        for subsystem_data in subsystem_data_list:
            subsystem_reference_time = None
            if "reference_time" in subsystem_data:
                subsystem_reference_time = subsystem_data["reference_time"]
            subsystem_data["csm"] = CoordinateSystemManager(
                subsystem_data["root_cs"],
                subsystem_data["name"],
                subsystem_reference_time,
            )
            subsystem_data["lcs"] = []

        cls._add_coordinate_systems_to_subsystems(tree, csm, subsystem_data_list)
        cls._merge_subsystems(tree, csm, subsystem_data_list)

        return csm



from dataclasses import dataclass, field

import numpy as np
import pandas as pd

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.asdf.validators import wx_shape_validator

__all__ = ["ShapeValidatorTestClass", "ShapeValidatorTestClassType"]


@dataclass
class ShapeValidatorTestClass:
    """Helper class to test the shape validator"""

    prop1: np.ndarray = np.ones((1, 2, 3))
    prop2: np.ndarray = np.ones((3, 2, 1))
    prop3: np.ndarray = np.ones((2, 4, 6, 8, 10))
    prop4: np.ndarray = np.ones((1, 3, 5, 7, 9))
    prop5: float = 3.141
    nested_prop: dict = field(
        default_factory=lambda: {
            "p1": np.ones((10, 8, 6, 4, 2)),
            "p2": np.ones((9, 7, 5, 3, 1)),
        }
    )
    time_prop: pd.DatetimeIndex = pd.timedelta_range("0s", freq="s", periods=9)
    optional_prop: np.ndarray = None


class ShapeValidatorTestClassType(WeldxType):
    """Helper class to test the shape validator"""

    name = "debug/test_shape_validator"
    version = "1.0.0"
    types = [ShapeValidatorTestClass]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {
        "wx_shape": wx_shape_validator,
    }

    @classmethod
    def to_tree(cls, node: ShapeValidatorTestClass, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = ShapeValidatorTestClass(**tree)
        return obj



import pandas as pd

import weldx

lcs = weldx.transformations.LocalCoordinateSystem(
    coordinates=[[1, 0, 0], [1, 2, 3]], time=pd.TimedeltaIndex(["0s", "2s"])
)
c = lcs.coordinates
c.weldx.time_ref = pd.Timestamp("2020-01-01 00:00:00")
print(c.time)
c.weldx.time_ref = pd.Timestamp("2020-01-01 00:00:01")
print(c.time)

buffer = weldx.asdf.utils._write_buffer({"LCS": lcs})
weldx.asdf.utils.notebook_fileprinter(buffer)



"""Examples for generating asdf schemas and classes."""

from weldx.asdf.utils import create_asdf_dataclass

create_asdf_dataclass(
    asdf_name="custom/testclass",
    asdf_version="1.0.0",
    class_name="TestClass",
    properties=[
        "prop1",
        "prop2",
        "prop3",
        "prop4",
        "list_prop",
        "pint_prop",
        "groove_prop",
    ],
    required=["prop1", "prop2", "prop3"],
    property_order=["prop1", "prop2", "prop3"],
    property_types=[
        "str",
        "int",
        "float",
        "bool",
        "List[str]",
        "pint.Quantity",
        "VGroove",
    ],
    description=[
        "a string",
        "",
        "a float",
        "a boolean value?",
        "a list",
        "a pint quantity",
        "a groove shape",
    ],
)



from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.measurement import DataTransformation

__all__ = ["DataTransformation", "DataTransformationType"]


class DataTransformationType(WeldxType):
    """Serialization class for data transformations."""

    name = "measurement/data_transformation"
    version = "1.0.0"
    types = [DataTransformation]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: DataTransformation, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = DataTransformation(**tree)
        return obj



import weldx.geometry as geo
import weldx.transformations as tf
from weldx import Q_
from weldx.asdf.tags.weldx.core.iso_groove import get_groove

linesegment = geo.LineSegment([[0, 0], [2, 1]])
print(linesegment.__repr__())
print()

arcsegment = geo.ArcSegment([[0, 1, 0], [1, 0, 0]], True)
print(arcsegment.__repr__())
print()

shape = geo.Shape([linesegment, arcsegment])
print(shape.__repr__())
print()

profile = geo.Profile(shape)
print(profile.__repr__())
print()

LHTS = geo.LinearHorizontalTraceSegment(5)
print(LHTS.__repr__())
print()

RHTS = geo.RadialHorizontalTraceSegment(3, 90)
print(RHTS.__repr__())
print()


# how to get this one running?
# varProfile = geo.VariableProfile([profile], [0], [1])
# print(varProfile.__repr__())
# print()

# LCS
LCS = tf.LocalCoordinateSystem(coordinates=[2, 4, -1])
print(LCS.__repr__())
print()

# create a linear trace segment
trace_segment = geo.LinearHorizontalTraceSegment(300)
trace = geo.Trace(trace_segment)
print(trace.__repr__())
print()

groove = get_groove(
    groove_type="VGroove",
    workpiece_thickness=Q_(0.5, "cm"),
    groove_angle=Q_(50, "deg"),
    root_face=Q_(1, "mm"),
    root_gap=Q_(1, "mm"),
)

# create 3d workpiece geometry from the groove profile and trace objects
geometry = geo.Geometry(groove.to_profile(width_default=Q_(4, "mm")), trace)

# Geometry - trace is empty -> no repr?
# geometry = geo.Geometry(profile, trace)
print(geometry.__repr__())
print()

# crete a new coordinate system manager with default base coordinate system
csm = tf.CoordinateSystemManager("base")

# add the workpiece coordinate system
csm.add_coordinate_system("workpiece", "base", trace.coordinate_system)

# CSM - need to lookup how to initialize this in the tutorial...
# CSM = tf.CoordinateSystemManager()
print(csm.__repr__())
print()



from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.measurement import GenericEquipment

__all__ = ["GenericEquipment", "GenericEquipmentType"]


class GenericEquipmentType(WeldxType):
    """Serialization class for generic-equipment."""

    name = "equipment/generic_equipment"
    version = "1.0.0"
    types = [GenericEquipment]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: GenericEquipment, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        if "sources" not in tree:
            tree["sources"] = None
        obj = GenericEquipment(**tree)
        return obj



"""Welding groove type definitions"""

from dataclasses import dataclass, field
from typing import List

import numpy as np
import pint

import weldx.geometry as geo
from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr
from weldx.asdf.validators import wx_unit_validator
from weldx.constants import WELDX_QUANTITY as Q_

_DEFAULT_LEN_UNIT = "mm"


def get_groove(
    groove_type,
    workpiece_thickness=None,
    workpiece_thickness2=None,
    root_gap=None,
    root_face=None,
    root_face2=None,
    root_face3=None,
    bevel_radius=None,
    bevel_radius2=None,
    bevel_angle=None,
    bevel_angle2=None,
    groove_angle=None,
    groove_angle2=None,
    special_depth=None,
    code_number=None,
):
    """Create a Groove from weldx.asdf.tags.weldx.core.groove.

    Make a selection from the given groove types.
    Groove Types:
        "VGroove"
        "UGroove"
        "IGroove"
        "UVGroove"
        "VVGroove"
        "HVGroove"
        "HUGroove"
        "DoubleVGroove"
        "DoubleUGroove"
        "DoubleHVGroove"
        "DoubleHUGroove"
        "FrontalFaceGroove"

    Each groove type has a different set of attributes which are required. Only
    required attributes are considered. All the required attributes for Grooves
    are in Quantity values from pint and related units are accepted.
    Required Groove attributes:
        "VGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            alpha: groove_angle
                - The groove angle is the whole angle of the V-Groove.
                  It is a pint Quantity in degree or radian.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the length of the Y-Groove which is not
                  part of the V. It can be 0.

        "UGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            beta: bevel_angle
                - The bevel angle is the angle that emerges from the circle segment.
                  Where 0 degree would be a parallel to the root face and 90 degree
                  would be a parallel to the workpiece.
            R: bevel_radius
                - The bevel radius defines the length of the radius of the U-segment.
                  It is usually 6 millimeters.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the height of the part below the U-segment.

        "IGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.

        "UVGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            alpha: groove_angle
                - The groove angle is the whole angle of the V-Groove part.
                  It is a pint Quantity in degree or radian.
            beta: bevel_angle
                - The bevel angle is the angle that emerges from the circle segment.
                  Where 0 degree would be a parallel to the root face and 90 degree
                  would be a parallel to the workpiece.
            R: bevel_radius
                - The bevel radius defines the length of the radius of the U-segment.
                  It is usually 6 millimeters.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            h: root_face
                - The root face is the height of the V-segment.

        "VVGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            alpha: groove_angle
                - The groove angle is the whole angle of the lower V-Groove part.
                  It is a pint Quantity in degree or radian.
            beta: bevel_angle
                - The bevel angle is the angle of the upper V-Groove part.
                  It is a pint Quantity in degree or radian.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the height of the part below the lower V-segment.
                  It can be 0 or None.
            h: root_face2
                - This root face is the height of the part of the lower V-segment
                  and the root face c.

        "HVGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            beta: bevel_angle
                - The bevel angle is the angle of the V-Groove part.
                  It is a pint Quantity in degree or radian.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the height of the part below the V-segment.

        "HUGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            beta: bevel_angle
                - The bevel angle is the angle that emerges from the circle segment.
                  Where 0 degree would be a parallel to the root face and 90 degree
                  would be a parallel to the workpiece.
            R: bevel_radius
                - The bevel radius defines the length of the radius of the U-segment.
                  It is usually 6 millimeters.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the height of the part below the U-segment.

        "DoubleVGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            alpha: groove_angle
                - The groove angle is the whole angle of the upper V-Groove part.
                  It is a pint Quantity in degree or radian.
            alpha2: groove_angle
                - The groove angle is the whole angle of the lower V-Groove part.
                  It is a pint Quantity in degree or radian.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the height of the part between the V-segments.
            h1: root_face2
                - The root face is the height of the upper V-segment.
                  Only c is needed.
            h2: root_face3
                - The root face is the height of the lower V-segment.
                  Only c is needed.

        "DoubleUGroove":
            t: workpiece_thickness
                - The workpiece thickness is a length Quantity, e.g.: "mm".
                  It is assumed that both workpieces have the same thickness.
            beta: bevel_angle
                - The bevel angle is the angle that emerges from the circle segment.
                  Where 0 degree would be a parallel to the root face and 90 degree
                  would be a parallel to the workpiece. The upper U-segment.
            beta2: bevel_angle2
                - The bevel angle is the angle that emerges from the circle segment.
                  Where 0 degree would be a parallel to the root face and 90 degree
                  would be a parallel to the workpiece. The lower U-segment.
            R: bevel_radius
                - The bevel radius defines the length of the radius of the
                  upper U-segment. It is usually 6 millimeters.
            R2: bevel_radius2
                - The bevel radius defines the length of the radius of the
                  lower U-segment. It is usually 6 millimeters.
            b: root_gap
                - The root gap is the distance of the 2 workpieces.
                  It can be 0 or None.
            c: root_face
                - The root face is the height of the part between the U-segments.
            h1: root_face2
                - The root face is the height of the upper U-segment.
                  Only c is needed.
            h2: root_face3
                - The root face is the height of the lower U-segment.
                  Only c is needed.

        "DoubleHVGroove":
            This is a special case of the DoubleVGroove. The values of the angles are
            interpreted here as bevel angel. So you have only half of the size.
            Accordingly the inputs beta1 (bevel angle) and beta2 (bevel angle 2)
            are used.

        "DoubleHUGroove":
            This is a special case of the DoubleUGroove.
            The parameters remain the same.

        "FrontalFaceGroove":
            These grooves are identified by their code number. These correspond to the
            key figure numbers from the standard. For more information, see the
            documentation.


    Example:
        from weldx import Q_
        from weldx.asdf.tags.weldx.core.groove import get_groove

        get_groove(groove_type="VGroove",
                   workpiece_thickness=Q_(9, "mm"),
                   groove_angle=Q_(50, "deg"),
                   root_face=Q_(4, "mm"),
                   root_gap=Q_(2, "mm"))

        get_groove(groove_type="UGroove",
                   workpiece_thickness=Q_(15, "mm"),
                   bevel_angle=Q_(9, "deg"),
                   bevel_radius=Q_(6, "mm"),
                   root_face=Q_(3, "mm"),
                   root_gap=Q_(1, "mm"))

    Parameters
    ----------
    groove_type :
        String specifying the Groove type
    workpiece_thickness :
        workpiece thickness (Default value = None)
    workpiece_thickness2 :
        workpiece thickness if type needs 2 thicknesses (Default value = None)
    root_gap :
        root gap, gap between work pieces (Default value = None)
    root_face :
        root face, upper part when 2 root faces are needed, middle part
        when 3 are needed (Default value = None)
    root_face2 :
        root face, the lower part when 2 root faces are needed. upper
        part when 3 are needed - used when min. 2 parts are needed
        (Default value = None)
    root_face3 :
        root face, usually the lower part - used when 3 parts are needed
        (Default value = None)
    bevel_radius :
        bevel radius (Default value = None)
    bevel_radius2 :
        bevel radius - lower radius for DU-Groove (Default value = None)
    bevel_angle :
        bevel angle, usually the upper angle (Default value = None)
    bevel_angle2 :
        bevel angle, usually the lower angle (Default value = None)
    groove_angle :
        groove angle, usually the upper angle (Default value = None)
    groove_angle2 :
        groove angle, usually the lower angle (Default value = None)
    special_depth :
        special depth used for 4.1.2 Frontal-Face-Groove (Default value = None)
    code_number :
        String, used to define the Frontal Face Groove (Default value = None)

    Returns
    -------
    type
        an Groove from weldx.asdf.tags.weldx.core.groove

    """
    if groove_type == "VGroove":
        return VGroove(
            t=workpiece_thickness, alpha=groove_angle, b=root_gap, c=root_face
        )

    elif groove_type == "UGroove":
        return UGroove(
            t=workpiece_thickness,
            beta=bevel_angle,
            R=bevel_radius,
            b=root_gap,
            c=root_face,
        )

    elif groove_type == "IGroove":
        return IGroove(t=workpiece_thickness, b=root_gap)

    elif groove_type == "UVGroove":
        return UVGroove(
            t=workpiece_thickness,
            alpha=groove_angle,
            beta=bevel_angle,
            R=bevel_radius,
            b=root_gap,
            h=root_face,
        )

    elif groove_type == "VVGroove":
        return VVGroove(
            t=workpiece_thickness,
            alpha=groove_angle,
            beta=bevel_angle,
            b=root_gap,
            c=root_face,
            h=root_face2,
        )

    elif groove_type == "HVGroove":
        return HVGroove(
            t=workpiece_thickness, beta=bevel_angle, b=root_gap, c=root_face
        )

    elif groove_type == "HUGroove":
        return HUGroove(
            t=workpiece_thickness,
            beta=bevel_angle,
            R=bevel_radius,
            b=root_gap,
            c=root_face,
        )

    elif groove_type == "DoubleVGroove":
        return DVGroove(
            t=workpiece_thickness,
            alpha_1=groove_angle,
            alpha_2=groove_angle2,
            b=root_gap,
            c=root_face,
            h1=root_face2,
            h2=root_face3,
        )

    elif groove_type == "DoubleUGroove":
        return DUGroove(
            t=workpiece_thickness,
            beta_1=bevel_angle,
            beta_2=bevel_angle2,
            R=bevel_radius,
            R2=bevel_radius2,
            b=root_gap,
            c=root_face,
            h1=root_face2,
            h2=root_face3,
        )

    elif groove_type == "DoubleHVGroove":
        return DHVGroove(
            t=workpiece_thickness,
            beta_1=bevel_angle,
            beta_2=bevel_angle2,
            c=root_face,
            h1=root_face2,
            h2=root_face3,
            b=root_gap,
        )

    elif groove_type == "DoubleHUGroove":
        return DHUGroove(
            t=workpiece_thickness,
            beta_1=bevel_angle,
            beta_2=bevel_angle2,
            R=bevel_radius,
            R2=bevel_radius2,
            b=root_gap,
            c=root_face,
            h1=root_face2,
            h2=root_face3,
        )

    elif groove_type == "FrontalFaceGroove":
        return FFGroove(
            t_1=workpiece_thickness,
            t_2=workpiece_thickness2,
            alpha=groove_angle,
            b=root_gap,
            e=special_depth,
            code_number=code_number,
        )

    else:
        raise ValueError(f"{groove_type} is not a groove type or not yet implemented.")


class BaseGroove:
    """Generic base class for all groove types."""

    def parameters(self):
        """Return groove parameters as dictionary of quantities."""
        return {k: v for k, v in self.__dict__.items() if isinstance(v, pint.Quantity)}

    def param_strings(self):
        """Generate string representation of parameters."""
        return [f"{k}={v:~}" for k, v in self.parameters().items()]

    def _ipython_display_(self):
        """Display the Groove as plot in notebooks."""
        self.plot()

    def plot(
        self,
        title=None,
        axis_label=None,
        raster_width=0.5,
        show_params=True,
        axis="equal",
        grid=True,
        line_style=".-",
        ax=None,
    ):
        """Plot a 2D-Profile.

        Parameters
        ----------
        title :
             (Default value = None)
        axis_label :
            label string to pass onto matplotlib (Default value = None)
        raster_width :
             (Default value = 0.1)
        show_params :
             (Default value = True)
        axis :
             (Default value = "equal")
        grid :
             (Default value = True)
        line_style :
             (Default value = ".")
        ax :
             (Default value = None)

       """
        profile = self.to_profile()
        if title is None:
            title = _groove_type_to_name[self.__class__]

        if show_params:
            title = title + "\n" + ", ".join(self.param_strings())

        profile.plot(
            title, raster_width, None, axis, axis_label, grid, line_style, ax=ax
        )

    def to_profile(self, width_default: pint.Quantity = None) -> geo.Profile:
        """Implements profile generation.

        Parameters
        ----------
        width_default :
             optional width to extend each side of the profile (Default value = None)

       """
        raise NotImplementedError("to_profile() must be defined in subclass.")


@dataclass
class VGroove(BaseGroove):
    """A Single-V Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    alpha :
        groove angle
    b :
        root gap
    c :
        root face
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    alpha: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.3", "1.5"])

    def to_profile(self, width_default=Q_(2, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(2, "mm"))


       """
        t = self.t  # .to(_DEFAULT_LEN_UNIT).magnitude
        alpha = self.alpha  # .to("rad").magnitude
        b = self.b  # .to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c  # .to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default  # .to(_DEFAULT_LEN_UNIT).magnitude

        # Calculations:
        s = np.tan(alpha / 2) * (t - c)

        # Scaling
        edge = np.append(-s, 0).min()
        if width <= -edge + Q_(1, "mm"):
            width = width - edge

        # Bottom segment
        x_value = np.append(-width, 0)
        y_value = Q_([0, 0], "mm")
        segment_list = ["line"]

        # root face
        if c != 0:
            x_value = np.append(x_value, 0)
            y_value = np.append(y_value, c)
            segment_list.append("line")

        # groove face
        x_value = np.append(x_value, -s)
        y_value = np.append(y_value, t)
        segment_list.append("line")

        # Top segment
        x_value = np.append(x_value, -width)
        y_value = np.append(y_value, t)
        segment_list.append("line")

        shape = _helperfunction(segment_list, [x_value, y_value])

        shape = shape.translate(np.append(-b / 2, 0))
        # y-axis is mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class VVGroove(BaseGroove):
    """A VV-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    alpha :
        groove angle
    beta :
        bevel angle
    b :
        root gap
    c :
        root face
    h :
        root face 2
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    alpha: pint.Quantity
    beta: pint.Quantity
    h: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.7"])

    def to_profile(self, width_default=Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        alpha = self.alpha.to("rad").magnitude
        beta = self.beta.to("rad").magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c.to(_DEFAULT_LEN_UNIT).magnitude
        h = self.h.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # Calculations
        h_lower = h - c
        h_upper = t - h
        s_1 = np.tan(alpha / 2) * h_lower
        s_2 = np.tan(beta) * h_upper

        # Scaling
        edge = np.min([-(s_1 + s_2), 0])
        if width <= -edge + 1:
            # adjustment of the width
            width = width - edge

        # x-values
        x_value = [-width, 0]
        # y-values
        y_value = [0, 0]
        segment_list = ["line"]

        if c != 0:
            x_value.append(0)
            y_value.append(c)
            segment_list.append("line")

        x_value += [-s_1, -s_1 - s_2, -width]
        y_value += [h + c, t, t]
        segment_list += ["line", "line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])

        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class UVGroove(BaseGroove):
    """A UV-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    alpha :
        groove angle
    beta :
        bevel angle
    R :
        bevel radius
    b :
        root gap
    h :
        root face
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    alpha: pint.Quantity
    beta: pint.Quantity
    R: pint.Quantity
    h: pint.Quantity = Q_(0, "mm")
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.6"])

    def to_profile(self, width_default: pint.Quantity = Q_(2, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(2, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        alpha = self.alpha.to("rad").magnitude
        beta = self.beta.to("rad").magnitude
        R = self.R.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        h = self.h.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # calculations:
        x_1 = np.tan(alpha / 2) * h
        # Center of the circle [0, y_m]
        y_circle = np.sqrt(R ** 2 - x_1 ** 2)
        y_m = h + y_circle
        # From next point to circle center is the vector (x,y)
        x = R * np.cos(beta)
        y = R * np.sin(beta)
        x_arc = -x
        y_arc = y_m - y
        # X-section of the upper edge
        x_end = x_arc - (t - y_arc) * np.tan(beta)

        # Scaling
        edge = np.max([-x_end, 0])
        if width <= edge + 1:
            # adjustment of the width
            width = width + edge

        # x-values
        x_value = [-width, 0, -x_1, 0, x_arc, x_end, -width]
        # y-values
        y_value = [0, 0, h, y_m, y_arc, t, t]
        segment_list = ["line", "line", "arc", "line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])

        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class UGroove(BaseGroove):
    """An U-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    beta :
        bevel angle
    R :
        bevel radius
    b :
        root gap
    c :
        root face
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    beta: pint.Quantity
    R: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.8"])

    def to_profile(self, width_default: pint.Quantity = Q_(3, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(3, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        beta = self.beta.to("rad").magnitude
        R = self.R.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # calculations:
        # From next point to circle center is the vector (x,y)
        x = R * np.cos(beta)
        y = R * np.sin(beta)
        # m = [0,c+R] circle center
        # => [-x,c+R-y] is the next point

        s = np.tan(beta) * (t - (c + R - y))

        # Scaling
        edge = np.max([x + s, 0])
        if width <= edge + 1:
            # adjustment of the width
            width = width + edge

        # x-values
        x_value = []
        # y-values
        y_value = []
        segment_list = []

        # bottom segment
        x_value.append(-width)
        y_value.append(0)
        x_value.append(0)
        y_value.append(0)
        segment_list.append("line")

        # root face
        if c != 0:
            x_value.append(0)
            y_value.append(c)
            segment_list.append("line")

        # groove face arc (circle center)
        x_value.append(0)
        y_value.append(c + R)

        # groove face arc
        x_value.append(-x)
        y_value.append(c + R - y)
        segment_list.append("arc")

        # groove face line
        x_value.append(-x - s)
        y_value.append(t)
        segment_list.append("line")

        # top segment
        x_value.append(-width)
        y_value.append(t)
        segment_list.append("line")

        shape = _helperfunction(segment_list, [x_value, y_value])

        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class IGroove(BaseGroove):
    """An I-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    b :
        root gap
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.2.1", "1.2.2", "2.1"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # x-values
        x_value = [-width, 0, 0, -width]
        # y-values
        y_value = [0, 0, t, t]
        segment_list = ["line", "line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])

        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class HVGroove(BaseGroove):
    """A HV-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    beta :
        bevel angle
    b :
        root gap
    c :
        root face
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    beta: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.9.1", "1.9.2", "2.8"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        beta = self.beta.to("rad").magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # Calculations
        s = np.tan(beta) * (t - c)

        # Scaling
        edge = np.min([-s, 0])
        if width <= -edge + 1:
            # adjustment of the width
            width = width - edge

        x_value = [-width, 0]
        y_value = [0, 0]
        segment_list = ["line"]

        if c != 0:
            x_value.append(0)
            y_value.append(c)
            segment_list.append("line")

        x_value += [-s, -width]
        y_value += [t, t]
        segment_list += ["line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])
        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        shape_h = geo.Shape()
        shape_h.add_line_segments(
            [[-width - (b / 2), 0], [-b / 2, 0], [-b / 2, t], [-width - (b / 2), t]]
        )

        return geo.Profile([shape_h, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class HUGroove(BaseGroove):
    """A HU-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    beta :
        bevel angle
    R :
        bevel radius
    b :
        root gap
    c :
        root face
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    beta: pint.Quantity
    R: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["1.11", "2.10"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        beta = self.beta.to("rad").magnitude
        R = self.R.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # Calculations
        x = R * np.cos(beta)
        y = R * np.sin(beta)
        s = np.tan(beta) * (t - (c + R - y))

        # Scaling
        edge = np.max([x + s, 0])
        if width <= edge + 1:
            # adjustment of the width
            width = width + edge

        x_value = [-width, 0]
        y_value = [0, 0]
        segment_list = ["line"]

        if c != 0:
            x_value.append(0)
            y_value.append(c)
            segment_list.append("line")

        x_value += [0, -x, -x - s, -width]
        y_value += [c + R, c + R - y, t, t]
        segment_list += ["arc", "line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])
        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        shape_h = geo.Shape()
        shape_h.add_line_segments(
            [[-width - (b / 2), 0], [-b / 2, 0], [-b / 2, t], [-width - (b / 2), t]]
        )

        return geo.Profile([shape_h, shape_r], units=_DEFAULT_LEN_UNIT)


# double Grooves
@dataclass
class DVGroove(BaseGroove):
    """A DV-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    alpha_1 :
        groove angle (upper)
    alpha_2 :
        groove angle (lower)
    b :
        root gap
    c :
        root face (middle)
    h1 :
        root face (upper)
    h2 :
        root face (lower)
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    alpha_1: pint.Quantity
    alpha_2: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    h1: pint.Quantity = None
    h2: pint.Quantity = None
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["2.4", "2.5.1", "2.5.2"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        alpha_1 = self.alpha_1.to("rad").magnitude
        alpha_2 = self.alpha_2.to("rad").magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c.to(_DEFAULT_LEN_UNIT).magnitude
        if self.h1 is None and self.h2 is None:
            h1 = (t - c) / 2
            h2 = (t - c) / 2
        elif self.h1 is not None and self.h2 is None:
            h1 = self.h1.to(_DEFAULT_LEN_UNIT).magnitude
            h2 = h1
        elif self.h1 is None and self.h2 is not None:
            h2 = self.h2.to(_DEFAULT_LEN_UNIT).magnitude
            h1 = h2
        else:
            h1 = self.h1.to(_DEFAULT_LEN_UNIT).magnitude
            h2 = self.h2.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # Calculations
        s_upper = np.tan(alpha_1 / 2) * h1
        s_lower = np.tan(alpha_2 / 2) * h2

        # Scaling
        edge = np.min([-s_upper, -s_lower, 0])
        if width <= -edge + 1:
            # adjustment of the width
            width = width - edge

        x_value = [-width, -s_lower, 0]
        y_value = [0, 0, h2]
        segment_list = ["line", "line"]

        if c != 0:
            x_value.append(0)
            y_value.append(h2 + c)
            segment_list.append("line")

        x_value += [-s_upper, -width]
        y_value += [t, t]
        segment_list += ["line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])
        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class DUGroove(BaseGroove):
    """A DU-Groove

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    beta_1 :
        bevel angle (upper)
    beta_2 :
        bevel angle (lower)
    R :
        bevel radius (upper)
    R2 :
        bevel radius (lower)
    b :
        root gap
    c :
        root face (middle)
    h1 :
        root face (upper)
    h2 :
        root face (lower)
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    beta_1: pint.Quantity
    beta_2: pint.Quantity
    R: pint.Quantity
    R2: pint.Quantity
    c: pint.Quantity = Q_(3, "mm")
    h1: pint.Quantity = None
    h2: pint.Quantity = None
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["2.7"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        beta_1 = self.beta_1.to("rad").magnitude
        beta_2 = self.beta_2.to("rad").magnitude
        R = self.R.to(_DEFAULT_LEN_UNIT).magnitude
        R2 = self.R2.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        c = self.c.to(_DEFAULT_LEN_UNIT).magnitude
        if self.h1 is None and self.h2 is None:
            h1 = (t - c) / 2
            h2 = (t - c) / 2
        elif self.h1 is not None and self.h2 is None:
            h1 = self.h1.to(_DEFAULT_LEN_UNIT).magnitude
            h2 = h1
        elif self.h1 is None and self.h2 is not None:
            h2 = self.h2.to(_DEFAULT_LEN_UNIT).magnitude
            h1 = h2
        else:
            h1 = self.h1.to(_DEFAULT_LEN_UNIT).magnitude
            h2 = self.h2.to(_DEFAULT_LEN_UNIT).magnitude
        width = width_default.to(_DEFAULT_LEN_UNIT).magnitude

        # Calculations
        x_upper = R * np.cos(beta_1)
        y_upper = R * np.sin(beta_1)
        s_upper = np.tan(beta_1) * (h1 - (R - y_upper))
        x_lower = R2 * np.cos(beta_2)
        y_lower = R2 * np.sin(beta_2)
        s_lower = np.tan(beta_2) * (h2 - (R2 - y_lower))

        # Scaling
        edge = np.max([x_upper + s_upper, x_lower + s_lower, 0])
        if width <= edge + 1:
            # adjustment of the width
            width = width + edge

        x_value = [-width, -(s_lower + x_lower), -x_lower, 0, 0]
        y_value = [0, 0, h2 - (R2 - y_lower), h2 - R2, h2]
        segment_list = ["line", "line", "arc"]

        if c != 0:
            x_value.append(0)
            y_value.append(h1 + c)
            segment_list.append("line")

        x_value += [0, -x_upper, -(s_upper + x_upper), -width]
        y_value += [h2 + c + R, t - (h1 - (R - y_upper)), t, t]
        segment_list += ["arc", "line", "line"]

        shape = _helperfunction(segment_list, [x_value, y_value])
        shape = shape.translate([-b / 2, 0])
        # y-axis as mirror axis
        shape_r = shape.reflect_across_line([0, 0], [0, 1])

        return geo.Profile([shape, shape_r], units=_DEFAULT_LEN_UNIT)


@dataclass
class DHVGroove(BaseGroove):
    """A DHV-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    beta_1 :
        bevel angle (upper)
    beta_2 :
        bevel angle (lower)
    b :
        root gap
    c :
        root face (middle)
    h1 :
        root face (upper)
    h2 :
        root face (lower)
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    beta_1: pint.Quantity
    beta_2: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    h1: pint.Quantity = None
    h2: pint.Quantity = None
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["2.9.1", "2.9.2"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        dv_groove = DVGroove(
            self.t,
            self.beta_1 * 2,
            self.beta_2 * 2,
            self.c,
            self.h1,
            self.h2,
            self.b,
            self.code_number,
        )
        dv_profile = dv_groove.to_profile(width_default)
        right_shape = dv_profile.shapes[1]

        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude
        left_shape = geo.Shape()
        left_shape.add_line_segments(
            [
                [-width_default - (b / 2), 0],
                [-b / 2, 0],
                [-b / 2, t],
                [-width_default - (b / 2), t],
            ]
        )

        return geo.Profile([left_shape, right_shape], units=_DEFAULT_LEN_UNIT)


@dataclass
class DHUGroove(BaseGroove):
    """A DHU-Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t :
        workpiece thickness
    beta_1 :
        bevel angle (upper)
    beta_2 :
        bevel angle (lower)
    R :
        bevel radius (upper)
    R2 :
        bevel radius (lower)
    b :
        root gap
    c :
        root face (middle)
    h1 :
        root face (upper)
    h2 :
        root face (lower)
    code_number :
        Numbers of the standard

    """

    t: pint.Quantity
    beta_1: pint.Quantity
    beta_2: pint.Quantity
    R: pint.Quantity
    R2: pint.Quantity
    c: pint.Quantity = Q_(0, "mm")
    h1: pint.Quantity = None
    h2: pint.Quantity = None
    b: pint.Quantity = Q_(0, "mm")
    code_number: List[str] = field(default_factory=lambda: ["2.11"])

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        du_profile = DUGroove(
            self.t,
            self.beta_1,
            self.beta_2,
            self.R,
            self.R2,
            self.c,
            self.h1,
            self.h2,
            self.b,
            self.code_number,
        ).to_profile(width_default)
        right_shape = du_profile.shapes[1]

        t = self.t.to(_DEFAULT_LEN_UNIT).magnitude
        b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
        width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude
        left_shape = geo.Shape()
        left_shape.add_line_segments(
            [
                [-width_default - (b / 2), 0],
                [-b / 2, 0],
                [-b / 2, t],
                [-width_default - (b / 2), t],
            ]
        )

        return geo.Profile([left_shape, right_shape], units=_DEFAULT_LEN_UNIT)


# Frontal Face - Groove
@dataclass
class FFGroove(BaseGroove):
    """A Frontal Face Groove.

    For a detailed description of the execution look in get_groove.

    Parameters
    ----------
    t_1 :
        workpiece thickness
    t_2 :
        workpiece thickness, if second thickness is needed
    alpha :
        groove angle
    b :
        root gap
    e :
        special depth
    code_number :
        Numbers of the standard

    """

    t_1: pint.Quantity
    t_2: pint.Quantity = None
    alpha: pint.Quantity = None
    # ["1.12", "1.13", "2.12", "3.1.1", "3.1.2", "3.1.3", "4.1.1", "4.1.2", "4.1.3"]
    code_number: str = None
    b: pint.Quantity = None
    e: pint.Quantity = None

    def to_profile(self, width_default: pint.Quantity = Q_(5, "mm")) -> geo.Profile:
        """Calculate a Profile.

        Parameters
        ----------
        width_default :
             pint.Quantity (Default value = Q_(5, "mm"))


        """
        if (
            self.code_number == "1.12"
            or self.code_number == "1.13"
            or self.code_number == "2.12"
        ):
            t_1 = self.t_1.to(_DEFAULT_LEN_UNIT).magnitude
            width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude
            shape1 = geo.Shape()
            shape1.add_line_segments(
                [
                    [0, 0],
                    [2 * width_default + t_1, 0],
                    [2 * width_default + t_1, t_1],
                    [0, t_1],
                    [0, 0],
                ]
            )
            shape2 = geo.Shape()
            shape2.add_line_segments(
                [
                    [width_default, 0],
                    [width_default + t_1, 0],
                    [width_default + t_1, -width_default],
                    [width_default, -width_default],
                    [width_default, 0],
                ]
            )
            return geo.Profile([shape1, shape2], units=_DEFAULT_LEN_UNIT)
        elif self.code_number == "3.1.1":
            t_1 = self.t_1.to(_DEFAULT_LEN_UNIT).magnitude
            t_2 = self.t_2.to(_DEFAULT_LEN_UNIT).magnitude
            alpha = self.alpha.to("rad").magnitude
            b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
            width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude

            if width_default < t_1 + 1:
                width_default = t_1 + width_default

            # x = t_1
            # y = 0

            x_1 = np.cos(alpha) * width_default
            y_1 = np.sin(alpha) * width_default

            x_2 = np.cos(alpha + np.pi / 2) * t_1
            y_2 = np.sin(alpha + np.pi / 2) * t_1

            x_3 = x_1 + x_2
            y_3 = y_1 + y_2

            shape1 = geo.Shape()
            shape1.add_line_segments(
                [[t_1 + x_1, y_1], [t_1, 0], [t_1 + x_2, y_2], [t_1 + x_3, y_3]]
            )
            shape2 = geo.Shape()
            shape2.add_line_segments(
                [[width_default, -b], [0, -b], [0, -t_2 - b], [width_default, -t_2 - b]]
            )
            return geo.Profile([shape1, shape2], units=_DEFAULT_LEN_UNIT)
        elif self.code_number == "3.1.2":
            t_1 = self.t_1.to(_DEFAULT_LEN_UNIT).magnitude
            t_2 = self.t_2.to(_DEFAULT_LEN_UNIT).magnitude
            b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
            width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude
            shape1 = geo.Shape()
            shape1.add_line_segments(
                [[0, 0], [width_default, 0], [width_default, t_1], [0, t_1]]
            )
            shape2 = geo.Shape()
            shape2.add_line_segments(
                [
                    [0, -b],
                    [2 * width_default, -b],
                    [2 * width_default, -t_2 - b],
                    [0, -t_2 - b],
                ]
            )
            return geo.Profile([shape1, shape2], units=_DEFAULT_LEN_UNIT)
        elif self.code_number == "3.1.3" or self.code_number == "4.1.1":
            t_1 = self.t_1.to(_DEFAULT_LEN_UNIT).magnitude
            t_2 = self.t_2.to(_DEFAULT_LEN_UNIT).magnitude
            alpha = self.alpha.to("rad").magnitude
            b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
            width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude

            x = np.sin(alpha + np.pi / 2) * b + b
            y = np.cos(alpha + np.pi / 2) * b

            x_1 = np.sin(alpha) * t_2 + x
            y_1 = np.cos(alpha) * t_2 + y

            x_2 = np.sin(alpha + np.pi / 2) * (b + width_default) + b
            y_2 = np.cos(alpha + np.pi / 2) * (b + width_default)

            x_3 = x_1 + x_2 - x
            y_3 = y_1 + y_2 - y

            shape1 = geo.Shape()
            shape1.add_line_segments(
                [[-width_default, 0], [0, 0], [0, t_1], [-width_default, t_1]]
            )
            shape2 = geo.Shape()
            shape2.add_line_segments([[x_3, y_3], [x_1, y_1], [x, y], [x_2, y_2]])
            return geo.Profile([shape1, shape2], units=_DEFAULT_LEN_UNIT)
        elif self.code_number == "4.1.2":
            t_1 = self.t_1.to(_DEFAULT_LEN_UNIT).magnitude
            t_2 = self.t_2.to(_DEFAULT_LEN_UNIT).magnitude
            alpha = self.alpha.to("rad").magnitude
            e = self.e.to(_DEFAULT_LEN_UNIT).magnitude
            width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude

            x_1 = np.sin(alpha) * e
            y_1 = np.cos(alpha) * e

            x_2 = np.sin(alpha + np.pi) * (t_2 - e)
            y_2 = np.cos(alpha + np.pi) * (t_2 - e)

            x_3 = x_2 + np.sin(alpha + np.pi / 2) * width_default
            y_3 = y_2 + np.cos(alpha + np.pi / 2) * width_default

            x_4 = x_1 + np.sin(alpha + np.pi / 2) * width_default
            y_4 = y_1 + np.cos(alpha + np.pi / 2) * width_default

            shape1 = geo.Shape()
            shape1.add_line_segments(
                [[-width_default, 0], [0, 0], [0, t_1], [-width_default, t_1]]
            )
            shape2 = geo.Shape()
            shape2.add_line_segments([[x_4, y_4], [x_1, y_1], [x_2, y_2], [x_3, y_3]])
            return geo.Profile([shape1, shape2], units=_DEFAULT_LEN_UNIT)
        elif self.code_number == "4.1.3":
            t_1 = self.t_1.to(_DEFAULT_LEN_UNIT).magnitude
            t_2 = self.t_2.to(_DEFAULT_LEN_UNIT).magnitude
            b = self.b.to(_DEFAULT_LEN_UNIT).magnitude
            width_default = width_default.to(_DEFAULT_LEN_UNIT).magnitude
            shape1 = geo.Shape()
            shape1.add_line_segments(
                [[0, width_default], [0, 0], [t_1, 0], [t_1, width_default]]
            )
            shape2 = geo.Shape()
            shape2.add_line_segments(
                [
                    [-width_default, -b],
                    [t_1 + width_default, -b],
                    [t_1 + width_default, -t_2 - b],
                    [-width_default, -t_2 - b],
                    [-width_default, -b],
                ]
            )
            return geo.Profile([shape1, shape2], units=_DEFAULT_LEN_UNIT)
        else:
            raise ValueError(
                "Wrong code_number. The Code Number has to be"
                " one of the following strings: "
                '"1.12", "1.13", "2.12", "3.1.1", "3.1.2",'
                ' "3.1.3", "4.1.1", "4.1.2", "4.1.3"'
            )


def _helperfunction(segment, array) -> geo.Shape:
    """Calculate a shape from input.
    Input segment of successive segments as strings.
    Input array of the points in the correct sequence. e.g.:
    array = [[x-values], [y-values]]

    Parameters
    ----------
    segment :
        list of String, segment names ("line", "arc")
    array :
        array of 2 array,
        first array are x-values
        second array are y-values

    Returns
    -------
    type
        geo.Shape

    """
    segment_list = []
    counter = 0
    for elem in segment:
        if elem == "line":
            seg = geo.LineSegment(
                np.vstack(
                    [array[0][counter : counter + 2], array[1][counter : counter + 2]]
                )
            )
            segment_list.append(seg)
            counter += 1
        if elem == "arc":
            arr0 = [
                # begin
                array[0][counter],
                # end
                array[0][counter + 2],
                # circle center
                array[0][counter + 1],
            ]
            arr1 = [
                # begin
                array[1][counter],
                # end
                array[1][counter + 2],
                # circle center
                array[1][counter + 1],
            ]
            seg = geo.ArcSegment([arr0, arr1], False)
            segment_list.append(seg)
            counter += 2

    return geo.Shape(segment_list)


_groove_type_to_name = {
    VGroove: "SingleVGroove",
    VVGroove: "VVGroove",
    UVGroove: "UVGroove",
    UGroove: "SingleUGroove",
    IGroove: "IGroove",
    HVGroove: "HVGroove",
    HUGroove: "HUGroove",
    DVGroove: "DoubleVGroove",
    DUGroove: "DoubleUGroove",
    DHVGroove: "DoubleHVGroove",
    DHUGroove: "DoubleHUGroove",
    FFGroove: "FrontalFaceGroove",
}

_groove_name_to_type = {v: k for k, v in _groove_type_to_name.items()}


class GrooveType(WeldxType):
    """ASDF Groove type."""

    name = "core/iso_groove"
    version = "1.0.0"
    types = [
        VGroove,
        VVGroove,
        UVGroove,
        UGroove,
        IGroove,
        HVGroove,
        HUGroove,
        DVGroove,
        DUGroove,
        DHVGroove,
        DHUGroove,
        FFGroove,
    ]
    requires = ["weldx"]
    handle_dynamic_subclasses = True
    validators = {"wx_unit": wx_unit_validator}

    @classmethod
    def to_tree(cls, node, ctx):
        """Convert to tagged tree and remove all None entries from node dictionary."""
        if isinstance(node, tuple(_groove_type_to_name.keys())):
            tree = dict(
                components=drop_none_attr(node), type=_groove_type_to_name[type(node)],
            )
            return tree
        else:  # pragma: no cover
            raise ValueError(
                f"Unknown groove type for object {node} with type {type(node)}"
            )

    @classmethod
    def from_tree(cls, tree, ctx):
        """Convert from tagged tree to a groove."""
        if tree["type"] in _groove_name_to_type:
            obj = _groove_name_to_type[tree["type"]](**tree["components"])
            return obj
        else:  # pragma: no cover
            raise ValueError(f"Unknown groove name {tree['type']}")


def _create_test_grooves():
    """Create dictionary with examples for all groove variations."""
    v_groove = get_groove(
        groove_type="VGroove",
        workpiece_thickness=Q_(9, "mm"),
        groove_angle=Q_(50, "deg"),
        root_face=Q_(4, "mm"),
        root_gap=Q_(2, "mm"),
    )
    u_groove = get_groove(
        groove_type="UGroove",
        workpiece_thickness=Q_(15, "mm"),
        bevel_angle=Q_(9, "deg"),
        bevel_radius=Q_(6, "mm"),
        root_face=Q_(3, "mm"),
        root_gap=Q_(1, "mm"),
    )
    i_groove = get_groove(
        groove_type="IGroove", workpiece_thickness=Q_(4, "mm"), root_gap=Q_(4, "mm")
    )
    uv_groove = get_groove(
        groove_type="UVGroove",
        workpiece_thickness=Q_(12, "mm"),
        groove_angle=Q_(60, "deg"),
        bevel_angle=Q_(11, "deg"),
        bevel_radius=Q_(6, "mm"),
        root_face=Q_(4, "mm"),
        root_gap=Q_(2, "mm"),
    )
    vv_groove = get_groove(
        groove_type="VVGroove",
        workpiece_thickness=Q_(12, "mm"),
        groove_angle=Q_(70, "deg"),
        bevel_angle=Q_(13, "deg"),
        root_gap=Q_(3, "mm"),
        root_face=Q_(1, "mm"),
        root_face2=Q_(5, "mm"),
    )
    hv_groove = get_groove(
        groove_type="HVGroove",
        workpiece_thickness=Q_(9, "mm"),
        bevel_angle=Q_(55, "deg"),
        root_gap=Q_(2, "mm"),
        root_face=Q_(1, "mm"),
    )
    hu_groove = get_groove(
        groove_type="HUGroove",
        workpiece_thickness=Q_(18, "mm"),
        bevel_angle=Q_(15, "deg"),
        bevel_radius=Q_(8, "mm"),
        root_gap=Q_(2, "mm"),
        root_face=Q_(3, "mm"),
    )
    dv_groove = get_groove(
        groove_type="DoubleVGroove",
        workpiece_thickness=Q_(19, "mm"),
        groove_angle=Q_(40, "deg"),
        groove_angle2=Q_(60, "deg"),
        root_gap=Q_(2, "mm"),
        root_face=Q_(5, "mm"),
        root_face2=Q_(7, "mm"),
        root_face3=Q_(7, "mm"),
    )
    dv_groove2 = get_groove(
        groove_type="DoubleVGroove",
        workpiece_thickness=Q_(19, "mm"),
        groove_angle=Q_(40, "deg"),
        groove_angle2=Q_(60, "deg"),
        root_gap=Q_(2, "mm"),
        root_face=Q_(5, "mm"),
    )
    dv_groove3 = get_groove(
        groove_type="DoubleVGroove",
        workpiece_thickness=Q_(19, "mm"),
        groove_angle=Q_(40, "deg"),
        groove_angle2=Q_(60, "deg"),
        root_gap=Q_(2, "mm"),
        root_face=Q_(5, "mm"),
        root_face3=Q_(7, "mm"),
    )
    # DU grooves
    du_groove = get_groove(
        groove_type="DoubleUGroove",
        workpiece_thickness=Q_(33, "mm"),
        bevel_angle=Q_(8, "deg"),
        bevel_angle2=Q_(12, "deg"),
        bevel_radius=Q_(6, "mm"),
        bevel_radius2=Q_(6, "mm"),
        root_face=Q_(3, "mm"),
        root_face2=Q_(15, "mm"),
        root_gap=Q_(2, "mm"),
    )
    du_groove2 = get_groove(
        groove_type="DoubleUGroove",
        workpiece_thickness=Q_(33, "mm"),
        bevel_angle=Q_(8, "deg"),
        bevel_angle2=Q_(12, "deg"),
        bevel_radius=Q_(6, "mm"),
        bevel_radius2=Q_(6, "mm"),
        root_face=Q_(3, "mm"),
        root_gap=Q_(2, "mm"),
    )
    du_groove3 = get_groove(
        groove_type="DoubleUGroove",
        workpiece_thickness=Q_(33, "mm"),
        bevel_angle=Q_(8, "deg"),
        bevel_angle2=Q_(12, "deg"),
        bevel_radius=Q_(6, "mm"),
        bevel_radius2=Q_(6, "mm"),
        root_face=Q_(3, "mm"),
        root_face2=Q_(15, "mm"),
        root_face3=Q_(15, "mm"),
        root_gap=Q_(2, "mm"),
    )
    du_groove4 = get_groove(
        groove_type="DoubleUGroove",
        workpiece_thickness=Q_(33, "mm"),
        bevel_angle=Q_(8, "deg"),
        bevel_angle2=Q_(12, "deg"),
        bevel_radius=Q_(6, "mm"),
        bevel_radius2=Q_(6, "mm"),
        root_face=Q_(3, "mm"),
        root_face3=Q_(15, "mm"),
        root_gap=Q_(2, "mm"),
    )
    dhv_groove = get_groove(
        groove_type="DoubleHVGroove",
        workpiece_thickness=Q_(11, "mm"),
        bevel_angle=Q_(35, "deg"),
        bevel_angle2=Q_(60, "deg"),
        root_face2=Q_(5, "mm"),
        root_face=Q_(1, "mm"),
        root_gap=Q_(3, "mm"),
    )
    dhu_groove = get_groove(
        groove_type="DoubleHUGroove",
        workpiece_thickness=Q_(32, "mm"),
        bevel_angle=Q_(10, "deg"),
        bevel_angle2=Q_(20, "deg"),
        bevel_radius=Q_(8, "mm"),
        bevel_radius2=Q_(8, "mm"),
        root_face2=Q_(15, "mm"),
        root_face=Q_(2, "mm"),
        root_gap=Q_(2, "mm"),
    )
    ff_groove0 = get_groove(
        groove_type="FrontalFaceGroove",
        workpiece_thickness=Q_(5, "mm"),
        code_number="1.12",
    )
    ff_groove1 = get_groove(
        groove_type="FrontalFaceGroove",
        workpiece_thickness=Q_(5, "mm"),
        workpiece_thickness2=Q_(7, "mm"),
        groove_angle=Q_(80, "deg"),
        root_gap=Q_(1, "mm"),
        code_number="3.1.1",
    )
    ff_groove2 = get_groove(
        groove_type="FrontalFaceGroove",
        workpiece_thickness=Q_(2, "mm"),
        workpiece_thickness2=Q_(5, "mm"),
        root_gap=Q_(1, "mm"),
        code_number="3.1.2",
    )
    ff_groove3 = get_groove(
        groove_type="FrontalFaceGroove",
        workpiece_thickness=Q_(2, "mm"),
        workpiece_thickness2=Q_(5, "mm"),
        groove_angle=Q_(80, "deg"),
        root_gap=Q_(1, "mm"),
        code_number="3.1.3",
    )
    ff_groove4 = get_groove(
        groove_type="FrontalFaceGroove",
        workpiece_thickness=Q_(2, "mm"),
        workpiece_thickness2=Q_(5, "mm"),
        groove_angle=Q_(80, "deg"),
        special_depth=Q_(4, "mm"),
        code_number="4.1.2",
    )
    ff_groove5 = get_groove(
        groove_type="FrontalFaceGroove",
        workpiece_thickness=Q_(2, "mm"),
        workpiece_thickness2=Q_(5, "mm"),
        root_gap=Q_(1, "mm"),
        code_number="4.1.3",
    )

    test_data = dict(
        v_groove=(v_groove, VGroove),
        u_groove=(u_groove, UGroove),
        i_groove=(i_groove, IGroove),
        uv_groove=(uv_groove, UVGroove),
        vv_groove=(vv_groove, VVGroove),
        hv_groove=(hv_groove, HVGroove),
        hu_groove=(hu_groove, HUGroove),
        dv_groove=(dv_groove, DVGroove),
        dv_groove2=(dv_groove2, DVGroove),
        dv_groove3=(dv_groove3, DVGroove),
        du_groove=(du_groove, DUGroove),
        du_groove2=(du_groove2, DUGroove),
        du_groove3=(du_groove3, DUGroove),
        du_groove4=(du_groove4, DUGroove),
        dhv_groove=(dhv_groove, DHVGroove),
        dhu_groove=(dhu_groove, DHUGroove),
        ff_groove0=(ff_groove0, FFGroove),
        ff_groove1=(ff_groove1, FFGroove),
        ff_groove2=(ff_groove2, FFGroove),
        ff_groove3=(ff_groove3, FFGroove),
        ff_groove4=(ff_groove4, FFGroove),
        ff_groove5=(ff_groove5, FFGroove),
    )

    return test_data



from io import BytesIO
from pathlib import Path

import asdf
import jinja2
import pandas as pd

from weldx.asdf.constants import SCHEMA_PATH
from weldx.asdf.extension import WeldxAsdfExtension, WeldxExtension


def drop_none_attr(node):
    """Utility function simplify :code:`to_tree` methods of  dataclass objects.

    The function requires the node to be convertible to dictionary via :code:`__dict__`.
    And can therefore be applied to all classes created using the @dataclass operator.
    The result is "clean" dictionary with all None entries removed.
    A simple :code:`to_tree` function could be implemented as such::

        def to_tree(cls, node, ctx):
            tree = drop_none_attr(node)
            return tree

    Parameters
    ----------
    node :
        node to write to asdf-tree

    Returns
    -------
    dict
        The node dictionary with None entries removed.

    """
    tree = {k: v for (k, v) in node.__dict__.items() if v is not None}
    return tree


# jinja template functions ---------------------------------------

_DTYPE_DICT = pd.DataFrame(
    data={
        "py_type": [
            "str",
            "float",
            "int",
            "bool",
            "pint.Quantity",
            "VGroove",
            "UGroove",
        ],
        "asdf_type": [
            "string",
            "number",
            "integer",
            "boolean",
            "tag:stsci.edu:asdf/unit/quantity-1.1.0",
            "tag:weldx.bam.de:weldx/core/iso_groove-1.0.0",
            "tag:weldx.bam.de:weldx/core/iso_groove-1.0.0",
        ],
    }
)


def _asdf_dtype(py_type):
    prefix = "type: "

    # strip list
    if py_type.startswith("List["):
        py_type = py_type[5:-1]

    lookup = _DTYPE_DICT.py_type.isin([py_type])
    if lookup.any():
        asdf_type = _DTYPE_DICT.loc[lookup].asdf_type.iloc[0]
        if ("tag:" in asdf_type) or ("-" in asdf_type):
            prefix = "$ref: "
        return prefix + asdf_type

    return f"{prefix}<TODO ASDF_TYPE({py_type})>"


_DEFAULT_ASDF_DESCRIPTION = "<TODO DESCRIPTION>"

_loader = jinja2.FileSystemLoader(
    searchpath=["./asdf/templates", "./weldx/asdf/templates"]
)
_env = jinja2.Environment(loader=_loader, autoescape=True)
_env.globals.update(zip=zip)
_env.globals.update(_asdf_dtype=_asdf_dtype)
_env.globals.update(str=str)


def make_asdf_schema_string(
    asdf_name,
    asdf_version,
    properties,
    description=None,
    property_types=None,
    required=None,
    property_order=None,
    additional_properties="false",
    schema_title=_DEFAULT_ASDF_DESCRIPTION,
    schema_description=_DEFAULT_ASDF_DESCRIPTION,
    flow_style="block",
):
    """Generate default ASDF schema.

    Parameters
    ----------
    asdf_name :

    asdf_version :

    properties :

    description :
         (Default value = None)
    property_types :
         (Default value = None)
    required :
         (Default value = None)
    property_order :
         (Default value = None)
    additional_properties :
         (Default value = "false")
    schema_title :
         (Default value = _DEFAULT_ASDF_DESCRIPTION)
    schema_description :
         (Default value = _DEFAULT_ASDF_DESCRIPTION)
    flow_style :
         (Default value = "block")

   """

    if property_types is None:
        property_types = ["NO_TYPE"] * len(properties)

    if description is None:
        description = [_DEFAULT_ASDF_DESCRIPTION] * len(properties)
    description = [_DEFAULT_ASDF_DESCRIPTION if d == "" else d for d in description]

    template_file = "asdf_dataclass.yaml.jinja"
    template = _env.get_template(template_file)

    output_text = template.render(
        asdf_name=asdf_name,
        asdf_version=asdf_version,
        properties=properties,
        description=description,
        property_types=property_types,
        required=required,
        property_order=property_order,
        additional_properties=additional_properties,
        schema_title=schema_title,
        schema_description=schema_description,
        flow_style=flow_style,
    )  # this is where to put args to the template renderer

    return output_text


def make_python_class_string(
    class_name, asdf_name, asdf_version, properties, property_types, required
):
    """Generate default python dataclass and ASDF Type.

    Parameters
    ----------
    class_name :

    asdf_name :

    asdf_version :

    properties :

    property_types :

    required :


   """

    template_file = "asdf_dataclass.py.jinja"
    template = _env.get_template(template_file)

    lib_imports = {
        dtype.split(".")[0] for dtype in property_types if len(dtype.split(".")) > 1
    }

    output_text = template.render(
        class_name=class_name,
        asdf_name=asdf_name,
        asdf_version=asdf_version,
        properties=properties,
        property_types=property_types,
        required=required,
        lib_imports=lib_imports,
    )

    return output_text


def create_asdf_dataclass(
    asdf_name,
    asdf_version,
    class_name,
    properties,
    property_types=None,
    description=None,
    required=None,
    property_order=None,
    schema_title=_DEFAULT_ASDF_DESCRIPTION,
    schema_description=_DEFAULT_ASDF_DESCRIPTION,
):
    """Generates a ASDF schema file with corresponding python class for simple dataclasses.

    Parameters
    ----------
    asdf_name :
        full schema name including prefixes
    asdf_version :
        schema version as string
    class_name :
        name of the Python class to generate
    properties :
        list of property names
    property_types :
        list with Python dtypes for each property (Default value = None)
    description :
        list of property descriptions (Default value = None)
    required :
        list of parameters that are set to required (Default value = None)
    property_order :
        asdf schema property order (Default value = None)
    schema_title :
        asdf schema title (Default value = _DEFAULT_ASDF_DESCRIPTION)
    schema_description :
        asdf schema description (Default value = _DEFAULT_ASDF_DESCRIPTION)

   """
    asdf_file_path = Path(
        SCHEMA_PATH + f"/weldx.bam.de/weldx/{asdf_name}-{asdf_version}.yaml"
    ).resolve()

    asdf_schema_string = make_asdf_schema_string(
        asdf_name=asdf_name,
        asdf_version=asdf_version,
        properties=properties,
        description=description,
        property_types=property_types,
        required=required,
        property_order=property_order,
        schema_title=schema_title,
        schema_description=schema_description,
    )
    asdf_file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(str(asdf_file_path), "w") as file:
        file.write(asdf_schema_string)

    python_file_path = Path(__file__ + f"/../tags/weldx/{asdf_name}.py").resolve()

    python_class_string = make_python_class_string(
        class_name=class_name,
        asdf_name=asdf_name,
        asdf_version=asdf_version,
        properties=properties,
        property_types=property_types,
        required=required,
    )
    python_file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(str(python_file_path), "w") as file:
        file.write(python_class_string)

    return asdf_file_path, python_file_path


def _write_buffer(tree: dict, asdffile_kwargs: dict = None, write_kwargs: dict = None):
    """Write ASDF file into buffer.

    Parameters
    ----------
    tree:
        Tree object to serialize.
    asdffile_kwargs
        Additional keywords to pass to asdf.AsdfFile()
    write_kwargs
        Additional keywords to pass to asdf.AsdfFile.write_to()
        Weldx-Extensions are always set.

    Returns
    -------
    BytesIO
        Bytes buffer of the ASDF file.
    """
    if asdffile_kwargs is None:
        asdffile_kwargs = {}
    if write_kwargs is None:
        write_kwargs = {}

    buff = BytesIO()
    with asdf.AsdfFile(
        tree, extensions=[WeldxExtension(), WeldxAsdfExtension()], **asdffile_kwargs
    ) as ff:
        ff.write_to(buff, **write_kwargs)
        buff.seek(0)
    return buff


def _read_buffer(buffer: BytesIO, open_kwargs: dict = None):
    """Read ASDF file contents from buffer instance.

    Parameters
    ----------
    buffer
        Buffer containing ASDF file contents
    open_kwargs
        Additional keywords to pass to asdf.AsdfFile.open()
        Extensions are always set, copy_arrays=True is set by default.

    Returns
    -------
    dict
        ASDF file tree.

    """
    if open_kwargs is None:
        open_kwargs = {"copy_arrays": True}

    buffer.seek(0)
    with asdf.open(
        buffer, extensions=[WeldxExtension(), WeldxAsdfExtension()], **open_kwargs,
    ) as af:
        data = af.tree
    return data


def _write_read_buffer(
    tree: dict, asdffile_kwargs=None, write_kwargs=None, open_kwargs=None
):
    """Perform a buffered write/read roundtrip of a tree using default ASDF settings.
    Parameters
    ----------
    tree
        Tree object to serialize.
    asdffile_kwargs
        Additional keywords to pass to asdf.AsdfFile()
    write_kwargs
        Additional keywords to pass to asdf.AsdfFile.write_to()
        Extensions are always set.
    open_kwargs
        Additional keywords to pass to asdf.AsdfFile.open()
        Extensions are always set, copy_arrays=True is set by default.
    Returns
    -------
    dict
    """
    buffer = _write_buffer(tree, asdffile_kwargs, write_kwargs)
    return _read_buffer(buffer, open_kwargs)


try:  # pragma: no cover
    import IPython
    from pygments import highlight
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import get_lexer_by_name, get_lexer_for_filename
except ImportError:  # pragma: no cover
    pass
else:  # pragma: no cover

    def notebook_fileprinter(file, style="YAML"):
        """Prints the code from file/BytesIO  to notebook cell with syntax highlighting.

        Parameters
        ----------
        file
            filename or BytesIO buffer
        style
            Syntax style to use

        """
        if isinstance(file, BytesIO):
            file.seek(0)
            code = file.read()
            lexer = get_lexer_by_name(style)
        else:
            with open(file, "rb") as f:
                code = f.read()
            if Path(file).suffix == ".asdf":
                lexer = get_lexer_by_name("YAML")
            else:
                lexer = get_lexer_for_filename(file)

        parts = code.partition(b"\n...")
        code = parts[0].decode("utf-8") + parts[1].decode("utf-8")

        formatter = HtmlFormatter()
        return IPython.display.HTML(
            '<style type="text/css">{}</style>{}'.format(
                formatter.get_style_defs(".highlight"),
                highlight(code, lexer, formatter),
            )
        )
